/*
**
** CONTOUR.C   Takes surface height data over a rectangular grid and
**             makes contour plots, intensity plots, and 3-D surface
**             plots.  Also analysizes contours for centroid position,
**             area, and volume.
**
** Will Menninger, 9/89, programmed in Turbo C 2.0
**                 Latest update:  1-12-93
**
**
** History starting with 2.24
**
** 2.24  5-25-94  Fixed grid plotting bug.  Plots grids even when
**                low on memory.
**
** 2.25  6-22-94  Fixed grey scale bug in 3D plotting.
**
** 2.26  6-23-94  ** Major enhancement:  Added multi-color layering
**                to 3D plotting.  Also added new overhead mode in
**                3D options.
**
** 2.27  6-25-94  Added file saving
**
** 2.28  7-20-94  Fixed a bug with the axes in overhead mode
**                on the 3D plotting. (see v 2.29)
**
** 2.29  3-07-95  Added a batch mode.
**                Added a PCX file save feature.
**                Correctly restores 50 line mode.
**                Really fixed a bug with the axes in 3D overhead mode.
**                Store pointers in multiple graph loads for faster re-loads.
**                Fixed grey scale bug in contour plotting. (3-28-95)
**
** 2.30  5-15-95  Save rotated image, sample image, flip y-values.
**                Fixed a bug in the inverse_image() function.
**
** 2.31  9-12-95  Fixed a bug in the sample save (it wouldn't save if no
**                data was loaded)
**
** 2.32  12-1-95  Made major ticks occur at either 1, 2, or 5 spacing.
**                Changed specification of X and Y ranges to Xmin,Xmax
**                and Ymin,Ymax rather than CX,CY,XRADIUS,YRADIUS.
**                Added capability to print to LPT and COM ports.
**
** 2.33  12-8-95  Prevent contour from reading in title strings that
**                are too long.
**
** 2.34  12-11-95 Reduced number of default ticks on legend for overhead
**                3D plots (6c).  Re-did the grid plotting/memory storage
**                to be more robust (it was causing memory faults).
**
** 2.35  12-12-95 Fixed a minor bug in where the major ticks start.
**
** 2.36  2-2-96   Multiple plot files now list the titles so you can
**                see which one you are picking.
**
** 2.37  5-17-96  More options when printing:  Postscript prints can
**                be appended to a file, plus they can be printed with
**                any margins (not just default centered margins).
**
** 2.38  6-7-96   Default contour spacing is affected by what area you
**                choose to plot.
**
** 2.39  7-29-96  Fixed bug in page eject selection (wasn't ejecting
**                when sending directly to the printer).
**
** 2.40  11-12-96 Added option to scrunch the titles closer-in on
**                countour graphs.
**
**
*/

#include "contour.h"

#define VERSION "2.40"

static char     exename[80];
static char    *borderptr;
static char    *scrcolors;
static char    *pathptr;

static char     program_name[]="CONTOUR PLOTTER " VERSION;
static char    *main_menu[]={
                    "Load data file",
                    "Display disk directory",
                    "Enter graph settings",
                    "Plot 2-D intensity map",
                    "Plot 2-D contour lines",
                    "Plot 3-D surface",
                    "Calculate contour centroids/areas/volumes",
                    "More...",
                    "Exit"
                    };
static char    *second_menu[]={
                    "Save data with graph settings",
                    "Save data rotated by 90 degrees",
                    "Mirror data about center y value",
                    "Save sample data file",
                    "Delete OUT*.PS",
                    "Inverse a camera image",
                    "",
                    "More...",
                    "Return to main menu"
                    };
static char    *default_menu[]={
                    "Change the current path",
                    "Change screen colors",
                    "Change path menu",
                    "","","","",
                    "Save defaults",
                    "Return to main menu"
                    };
static char    *graph_menu[]={
                    "Enter graph title",
                    "Enter X-axis title",
                    "Enter Y-axis title",
                    "Enter Z-axis title",
                    "Enter graph size, axis margins",
                    "Enter page margins",
                    "Enter X-min,X-max",
                    "Enter Y-min,Y-max",
                    "Return to main menu"
                    };

static char     path[80];
static char    *paths[10];
static char     defpaths[]=
                   "\0"
                   "\0"
                   "\0"
                   "\0"
                   "\0"
                   "\0"
                   "\0"
                   "\0"
                   "\0";


static void    catalog       (void);
static STATUS  second_set    (DATAPTR dp);
static STATUS  defaults      (DATAPTR dp);
static void    inverse_image (DATAPTR dp);
static void    mirror_data   (DATAPTR dp);
static STATUS  change_graph  (DATAPTR dp);
static STATUS  init          (void);
static void    terminate     (DATAPTR d);
static void    setvmode      (int mode);


main(int argc,char *argv[])

    {
    DATA     d;
    int      m;
    static int first=1;
    char    buf[100];

    if (argc>1)
        setbatch(argv[1]);
    d.memory=NULL;
    d.pagex=-1;
    d.pagey=-1;
    d.pageleftmar=-99.;
    d.pagebotmar=-99.;
    d.no_title_margin=0;
    if (argc>0)
        strcpy(exename,argv[0]);
    if (init()==ERROR)
        return(10);
    init_grids();
    do
        {
        clrscrn();
        border_set(borderptr[0]);
        if (first)
            {
            first=0;
            gotoxy(1,24);
            hilite("");
            cprintf("Mem = %u, Far = %ld",coreleft(),farcoreleft());
            gotoxy(1,1);
            }
        hilite(program_name);
        normal("\r\n\r\nCurrent path:  ");
        hilite(path);
        normal("\r\nResident data:  ");
        if (loaded(0))
            {
            prshotname();
            sprintf(buf," (%d x %d)",d.xsize,d.ysize);
            hilite(buf);
            }
        else
            error("None loaded.");
        normal("\r\n\r\n");
        m=bmenu(main_menu);
        switch(m)
            {
            case 1:
                if (loadshot(path,&d)==NO_ERROR)
                    init_ranges(&d);
                break;
            case 2:
                catalog();
                break;
            case 3:
                change_graph(&d);
                break;
            case 4:
                iplot(&d);
                break;
            case 5:
                contour_plot(&d);
                break;
            case 6:
                do_3d(&d);
                break;
            case 7:
                area_stuff();
                break;
            case 8:
                second_set(&d);
                break;
            }
        } while (m!=9);
    terminate(&d);
    free_grids();
    killbatch();
    restore_text_mode(0);
    return(0);
    }


static STATUS second_set(DATAPTR dp)

    {
    int     m;

    do
        {
        clrscrn();
        border_set(borderptr[0]);
        hilite("IMAGE MANIPULATION OPTIONS");
        normal("\r\n\r\n");
        m=bmenu(second_menu);
        if (m==ESCAPE)
           return(ESCAPE);
        switch(m)
            {
            case 1:
                if (loaded(1))
                    saveshot(path,dp,0);
                break;
            case 2:
                if (loaded(1))
                    saveshot(path,dp,1);
                break;
            case 3:
                if (loaded(1))
                    mirror_data(dp);
                break;
            case 4:
                saveshot(path,dp,2);
                break;
            case 5:
                ps_delete();
                break;
            case 6:
                if (loaded(0))
                    inverse_image(dp);
                break;
            case 8:
                defaults(dp);
                break;
            }
        } while (m!=9);
    return(NO_ERROR);
    }


static STATUS defaults(DATAPTR dp)

    {
    int     m;

    do
        {
        clrscrn();
        border_set(borderptr[0]);
        hilite("OTHER OPTIONS");
        normal("\r\n\r\n");
        m=bmenu(default_menu);
        if (m==ESCAPE)
           return(ESCAPE);
        switch(m)
            {
            case 1:
                path_change(path,&paths[1]);
                break;
            case 2:
                changescrattrs(scrcolors,borderptr);
                break;
            case 3:
                defpathchange(pathptr,paths);
                break;
            case 8:
                savedefaults(exename);
                break;
            }
        } while (m!=9);
    return(NO_ERROR);
    }


static void inverse_image(DATAPTR dp)

    {
    int     i,j;

    normal("\r\n\r\n");
    hilite("Inversing entire data array...");
    for (i=0;i<dp->xsize;i++)
        for (j=0;j<dp->ysize;j++)
            dp->shotdata[i][j]=255.-dp->shotdata[i][j];
    for (i=0;i<dp->xsize-1;i++)
        for (j=0;j<dp->ysize-1;j++)
            dp->interp[i][j]=255.-dp->interp[i][j];
    }


static void mirror_data(DATAPTR dp)

    {
    int     i,j;
    float   f;

    normal("\r\n\r\n");
    hilite("Mirroring entire data array...");
    for (i=0;i<dp->xsize;i++)
        for (j=0;j<(dp->ysize>>1);j++)
            {
            f=dp->shotdata[i][j];
            dp->shotdata[i][j] = dp->shotdata[i][dp->ysize-j-1];
            dp->shotdata[i][dp->ysize-j-1] = f;
            }
    for (i=0;i<dp->xsize-1;i++)
        for (j=0;j<((dp->ysize-1)>>1);j++)
            {
            f=dp->interp[i][j];
            dp->interp[i][j] = dp->interp[i][dp->ysize-j-2];
            dp->interp[i][dp->ysize-j-2] = f;
            }
    }


static STATUS change_graph(DATAPTR dp)

    {
    int     m;
    char    buf[150];
    float   x,y;

    if (!loaded(1))
        return(NO_ERROR);
    do
        {
        clrscrn();
        border_set(borderptr[0]);
        hilite("GRAPH OPTIONS");
        normal("\r\n\r\n");
        sprintf(buf,"Title:   %s\r\n",dp->gtitle);
        normal(buf);
        sprintf(buf,"X Title: %s\r\n",dp->xtitle);
        normal(buf);
        sprintf(buf,"Y Title: %s\r\n",dp->ytitle);
        normal(buf);
        sprintf(buf,"Z Title: %s\r\n",dp->ztitle);
        normal(buf);
        {
        float   xp,yp;
        if (dp->pagex<0)
            xp=yp=6.0;
        else
            {
            xp=dp->pagex;
            yp=dp->pagey;
            }
        sprintf(buf,"Page:    %4.2f x %4.2f in (LF MAR=%4.2f, BOT MAR=%4.2f)\r\n",
                xp,yp,dp->pageleftmar,dp->pagebotmar);
        }
        normal(buf);
        if (dp->no_title_margin)
            {
            sprintf(buf,"Title close to graph box\r\n");
            normal(buf);
            }
        sprintf(buf,"Xmn,mx:  %g,%g\r\n",dp->xmin,dp->xmax);
        normal(buf);
        sprintf(buf,"Ymn,mx:  %g,%g\r\n\r\n",dp->ymin,dp->ymax);
        normal(buf);
        m=bmenu(graph_menu);
        if (m==ESCAPE)
            return(ESCAPE);
        switch(m)
            {
            case 1:
                if (bgetinput("Title:  ",buf,70)==ESCAPE)
                    break;
                strcpy(dp->gtitle,buf);
                break;
            case 2:
                if (bgetinput("X title: ",buf,70)==ESCAPE)
                    break;
                strcpy(dp->xtitle,buf);
                break;
            case 3:
                if (bgetinput("Y title: ",buf,70)==ESCAPE)
                    break;
                strcpy(dp->ytitle,buf);
                break;
            case 4:
                if (bgetinput("Z title: ",buf,70)==ESCAPE)
                    break;
                strcpy(dp->ztitle,buf);
                break;
            case 5:
                if (bgetinput("Page dimensions in inches (X,Y):  ",buf,30)
                    ==ESCAPE)
                    break;
                if (sscanf(buf,"%f,%f",&x,&y)<2)
                    break;
                if (x<1 || x>15 || y<1 || y>15)
                    {
                    error("Page dimensions out of range.");
                    normal("\r\n\r\n");
                    wait_enter();
                    break;
                    }
                dp->pagex=x;
                dp->pagey=y;
                if (bgetinput("Keep title close to graph box? (y/n):  ",buf,5)
                    ==ESCAPE)
                    break;
                dp->no_title_margin=(buf[0]=='y' || buf[0]=='Y');
                break;
            case 6:
                if (bgetinput("Page margins in inches (-99=CENTERED) (LEFT,BOTTOM):  ",buf,30)
                    ==ESCAPE)
                    break;
                if (sscanf(buf,"%f,%f",&x,&y)<2)
                    break;
                if (x<-115 || x>15 || y<-115 || y>15)
                    {
                    error("Page margins out of range.");
                    normal("\r\n\r\n");
                    wait_enter();
                    break;
                    }
                dp->pageleftmar=x;
                dp->pagebotmar=y;
                break;
            case 7:
                if (bgetinput("Enter Xmin,Xmax:  ",buf,30)==ESCAPE)
                    break;
                if (sscanf(buf,"%f,%f",&x,&y)<2)
                    break;
                if (x==y)
                    {
                    error("Xmin must not equal Xmax.");
                    normal("\r\n\r\n");
                    wait_enter();
                    break;
                    }
                dp->xmin=x;
                dp->xmax=y;
                break;
            case 8:
                if (bgetinput("Enter Ymin,Ymax:  ",buf,30)==ESCAPE)
                    break;
                if (sscanf(buf,"%f,%f",&x,&y)<2)
                    break;
                if (x==y)
                    {
                    error("Ymin must not equal Ymax.");
                    normal("\r\n\r\n");
                    wait_enter();
                    break;
                    }
                dp->ymin=x;
                dp->ymax=y;
                break;
            }
        } while (m!=9);
    return(NO_ERROR);
    }


static void catalog(void)

    {
    char    buf[20];
    char    wildspec[100];

    if (bgetinput("Enter wildcard [Def = *.*]:  ",buf,18)==ESCAPE)
        return;
    if (buf[0]==NULL)
        strcpy(buf,"*.*");
    strcpy(wildspec,path);
    strcat(wildspec,buf);
    directory(wildspec);
    }


static STATUS init(void)

    {
    if (no_graphics())
        return(ERROR);
    read_text_mode();
    setvmode(3);
    setlegalmode();
    cursoroff();
    borderptr=defptr();
    scrcolors=&borderptr[defscrattrs(borderptr)];
    pathptr=&borderptr[20];
    initpaths(pathptr,paths,defpaths);
    strcpy(path,paths[0]);
    strupr(path);
    window(1,1,25,80);
    clrscrn();
    border_set(borderptr[0]);
    return(NO_ERROR);
    }


static void terminate(DATAPTR d)

    {
    int y;

    if (d->memory!=NULL)
        {
        farfree(d->memory);
        d->memory=NULL;
        }
    y=wherey();
    window(1,y,80,25);
    textattr(7);
    clrscr();
    window(1,1,80,25);
    gotoxy(1,y);
    }


STATUS aborted(void)

    {
    cprintf("\r\nOperation aborted.\r\n");
    return(ERROR);
    }


void debug(char *s)

    {
    int     x,y;

    x=wherex();y=wherey();
    gotoxy(1,25);
    error("");
    printcm(' ',80);
    cputs(s);
    gotoxy(x,y);
    normal("");
    }


static void setvmode(int mode)

    {
    union REGS  set;

    set.h.ah=0;
    set.h.al=mode;
    int86(0x10,&set,&set);
    }
