/*
**
** PS3D.C      Plot 3D data to PostScript file
**
** Will Menninger, 10/89, programmed in Turbo C 2.0
**
*/

#include "contour.h"
#include <ps.h>

#define LTHICK      .48
#define AXLW        .48
/* times AXMAR * ps72 */
#define TKMAJ       0.6
/* times AXMAR * ps72 */
#define TKMIN       0.3
/* times ps72 */
#define AXMAR       0.02
/* times fontsize */
#define YAXNMAR     3.
/* times fontsize */
#define XAXNMAR     .5
#define TMAR        2.5
#define BOXLWIDTH   .72
#define VFRACB      .65
/* times pagewidth */
#define LEGX        0.05
/* times font size */
#define LEGMAR      3.
#define PI          3.1415926536
#define NBOXES      10

/* 3D layout */
#define FW      0.35
#define FH      0.35

/*
** These values are fractions of the graph width and height reserved
** for the 3-D perspective for viewtype=='a'
**

               <-------------pagex ---+---------->
                        --------------------------  ^
                       /|                        |  |
                      / |                        |  |
                     /  |                        |  |
                    /   |                        |  |
                   /    |                        |  |
                  /     |                        |  |
                 /      |                        |  |
                /       |                        |  |
               |        |                        |  |
               |        |                        |  |
               |<--FH-->|                        |  |
               |        |                        |  pagey
               |        |                        |  |
               |        --------------------------  |
               |       /       ^                /   |
               |      /        |               /    |
               |     /         |              /     |
               |    /         FV             /      |
               |   /           |            /       |
               |  /            |           /        |
               | /             |          /         |
               |/              V         /          |
               -------------------------/           V


** For viewtype=='b'


                 <-----------pagex -------------->
                                ---                 ^
                             --- | ---              |
                          ---    |    ---           |
                       ---       |       ---        |
                    ---          |          ---     |
                 ---             |             ---  |
                 |               |               |  |
                 |               |               |  |
                 |              ---              |  |
                 |           --- ^ ---           |  |
                 |        ---    |    ---        |  |
                 |     ---       |       ---     |  |
                 |  ---          |          ---  |  pagey
                 ---                           ---  |
                 ---            VFRACB         ---  |
                    ---          |          ---     |
                       ---       |       ---        |
                          ---    |    ---           |
                             --- V ---              |
                                ---                 V
                                 <------- 0.5 ------>

** For viewtype=='c'


      <--------------- pagex -------------->
      LEGX
      <--><------>
       --  LEGMAR -------------------------   ^
      |  |       |                         |  |
      |  |       |                         |  |
      |  |       |                         |  |
      |  |       |                         |  |
      |  |       |                         |  |
      |  |       |                         |  |
      |  |       |                         |  |
      |  |       |                         |  pagey
      |  |       |                         |  |
      |  |       |                         |  |
      |  |       |                         |  |
      |  |       |                         |  |
      |  |       |                         |  |
      |  |       |                         |  |
      |  |       |                         |  |
       --         -------------------------   V



**
*/

/* grey scales */
#define TOPLN   0.5
#define TOPFILL 0.9
#define BOTLN   0.
#define BOTFILL 0.65

/* colors */
#define CTOPLN      0.,1.,0.
#define CTOPFILL    0.,.5,.25
#define CBOTLN      1.,0.,0.
#define CBOTFILL    .5,.25,.25

#define AXES    1
#define AXESBRIGHT  15
#define TOPBRT  10
#define TOPDIM  2
#define BOTBRT  12
#define BOTDIM  4
#define BACKGRND 0
#define VIEW        1
#define CAPTURE     2

#define COLORSOH    8
#define COLORS3D    6
#define GREYS3D     4
#define GREYSOH     4

static float paletteOHc[COLORSOH][3]=
        {
        {0.,1.,0.},
        {0.,1.,1.},
        {0.,.6,1.},
        {.6,0.,1.},
        {1.,0.,1.},
        {1.,.6,0.},
        {.7,.4,0.},
        {.5,.5,.5},
        };
static float paletteOHg[GREYSOH]= { .9,.6,.4,.2 };
static float palette3Dc[COLORS3D][3]=
        {
        {0.,1.,0.}, /* green  */
        {1.,1.,0.}, /* yellow */
        {1.,.7,0.}, /* orange */
        {1.,0.,0.}, /* red    */
        {.7,0.,1.}, /* purple */
        {0.,.6,1.}  /* blue   */
        };
static float palette3Dg[GREYS3D]= { .9,.6,.4,.2 };


static float    ps72;
static float    psleft,psright,pstop,psbottom;
static float    psx0;
static float    psy0;
static float    lx0;
static float    gx0,gxx,gxy;
static float    gy0,gyx,gyy,gyz;
static int      xoff;
static int      yoff;
static float    zoff;
static int      xscale;
static int      yscale;
static float    zscale;
static float    theta;
static float    thrad;
static float    pagex,pagey;
static int      pscolor;
static float   *zboundary;
static int      nlayers;
static int      startlayer;
static int      viewtype;
static int      append,eject;

/* PostScript */

static float    fs;

static void triangle3d(float x1,float y1,float z1,float x2,float y2,
                       float z2,float x3,float y3,float z3);
static void tile3d(float x1,float y1,float z1,float x2,float y2,float z2,
                   float x3,float y3,float z3,float x4,float y4,float z4);
static void setup3d(DATAPTR dp);
static float x3d(float x,float y);
static float y3d(float x,float y,float z);
static void line3d(float x1,float y1,float z1,float x2,float y2,float z2);
static void setupaxes(DATAPTR d,int grids);
static void set_pscolor(int layer,int type);
static void poly_type(int layer,int npoints,int top);
static void def_colors(void);
static void three_verts(int grids,float z0,float z1);
static void z_legend(DATAPTR d);


void ps_3d(char *filename,DATAPTR dp,int xo,int xs,int yo,int ys,
           float zo,float zs,int vt,int grids,int color,
           float *zb,int nl,int sl,int _append,int _eject)

    {
    char    buf[100];
    int     x,y;
    float   fy;
    float   far*far*sd;
    float   far*far*in;

    append=_append;
    eject=_eject;
    zboundary=zb;
    nlayers=nl;
    startlayer=sl;
    pscolor=color;
    viewtype=vt;
    if (dp->pagex<0)
        {
        pagex=6.0;
        pagey=6.0;
        }
    else
        {
        pagex=dp->pagex;
        pagey=dp->pagey;
        }
    ps72=(pagex+pagey)*36.;
    sd=dp->shotdata;
    in=dp->interp;
    ps_update(0,15,"Writing PostScript to %s ...",filename);
    fs=fontsize(pagex,pagey);
    xoff=xo;
    xscale=xs;
    yoff=yo;
    yscale=ys;
    zoff=zo;
    zscale=zs;
    if (append)
        g_append(filename);
    else
        g_init(filename);
    sprintf(buf,"\\fTR\\s{%g}",fs);
    g_showtext(buf);

    setup3d(dp);
    setupaxes(dp,grids);

    g_linecap(1);
    g_linejoin(1);
    g_scale(.1,.1);
    g_linewidth( (xscale>25 ||  yscale>25) ? LTHICK*5. : LTHICK*10.);

    if (xscale>25 || yscale>25)
        for (y=yoff+yscale-1;y>=yoff;y--)
            {
            for (x=xoff;x<xoff+xscale;x++)
                tile3d(x,y,sd[x][y],x,y+1,sd[x][y+1],x+1,y+1,sd[x+1][y+1],
                       x+1,y,sd[x+1][y]);
            if (bbioskey(1))
                break;
            if (y%5==0)
                {
                fy=dp->ymin+y*(dp->ymax-dp->ymin)/(dp->ysize-1.);
                ps_update(0,15,"Y = %g",&fy);
                }
            }
    else
        for (y=yoff+yscale-1;y>=yoff;y--)
            {
            if (bbioskey(1))
                break;
            for (x=xoff;x<xoff+xscale;x++)
                triangle3d(x+.5,y+.5,in[x][y],x,y+1,sd[x][y+1],x+1,y+1,sd[x+1][y+1]);
            if (bbioskey(1))
                break;
            for (x=xoff;x<xoff+xscale;x++)
                {
                triangle3d(x+1,y,sd[x+1][y],x+.5,y+.5,in[x][y],x+1,y+1,sd[x+1][y+1]);
                triangle3d(x,y,sd[x][y],x,y+1,sd[x][y+1],x+.5,y+.5,in[x][y]);
                }
            if (bbioskey(1))
                break;
            for (x=xoff;x<xoff+xscale;x++)
                triangle3d(x+1,y,sd[x+1][y],x,y,sd[x][y],x+.5,y+.5,in[x][y]);
            if (y%5==0)
                {
                fy=dp->ymin+y*(dp->ymax-dp->ymin)/(dp->ysize-1.);
                ps_update(0,15,"Y = %g",&fy);
                }
            }
    g_unscale();
    if (viewtype=='c')
        {
        ps_bbox();
        ps_putgrid(grids);
        }
    if (eject)
        g_showpage();
    g_close();
    if (bbioskey(1))
        {
        bbioskey(0);
        ps_update(1,12,"PostScript printing interrupted.",NULL);
        finalprint(filename,0);
        }
    else
        finalprint(filename,1);
    }


static void setup3d(DATAPTR dp)

    {
    float   totalheight,totalwidth;
    float   xsquish,ysquish;
    char    buf[100];
    int     i,j;

    thrad=atan2(FH*pagey,FW*pagex);
    theta=thrad*57.29577;
    xsquish=cos(thrad);
    ysquish=sin(thrad);
    totalheight=pagey*72.+(1.+TMAR)*fs+(2.+XAXNMAR)*fs*ysquish
                 +ps72*AXMAR*ysquish;
    totalwidth=pagex*72.+(YAXNMAR+1.)*fs*xsquish+ps72*AXMAR;
    psbottom=dp->pagebotmar<-90. ? 11.*36.-totalheight/2. : dp->pagebotmar*72.;
    pstop=psbottom+totalheight;
    psleft=dp->pageleftmar<-90. ? 8.5*36.-totalwidth/2. : dp->pageleftmar*72.;
    psright=psleft+totalwidth;
    psx0=psright-pagex*72.;
    psy0=psbottom+ps72*AXMAR*ysquish+(2.+XAXNMAR)*fs*ysquish;
    if (viewtype=='a')
        {
        gxx=(1-FW)*pagex*72./xscale;
        gyy=FH*pagey*72./yscale;
        gxy=FW*pagex*72./yscale;
        gyz=(1-FH)*pagey*72./zscale;
        gyx=0.;
        gx0=psx0-xoff*gxx-yoff*gxy;
        gy0=psy0-xoff*gyx-yoff*gyy-zoff*gyz;
        thrad=atan2(FH*pagey,FW*pagex);
        }
    else if (viewtype=='b')
        {
        gxx=pagex*72./xscale/2.;
        gxy=pagex*72./yscale/2.;
        gx0=psx0-xoff*gxx-yoff*gxy;
        gyx=-pagey*72.*VFRACB/2./xscale;
        gyy=pagey*72.*VFRACB/2./yscale;
        gyz=pagey*72.*(1.-VFRACB)/zscale;
        gy0=psy0+pagey*72.*VFRACB/2.-xoff*gyx-yoff*gyy-zoff*gyz;
        thrad=atan2(VFRACB*pagey/2.,pagex/2.);
        }
    else
        {
        lx0 = pagex*72*LEGX + (LEGMAR+1+YAXNMAR)*fs + AXMAR*ps72;
        gxx=(pagex*72.-lx0)/xscale;
        gyy=pagey*72./yscale;
        gxy=0.;
        gyz=0.;
        gyx=0.;
        gx0=psx0+lx0-xoff*gxx;
        gy0=psy0-yoff*gyy;
        thrad=3.1415926536/2.;
        ps_setgriddim(dp->xmin+xoff*(dp->xmax-dp->xmin)/(dp->xsize-1.),
                      dp->ymin+yoff*(dp->ymax-dp->ymin)/(dp->ysize-1.),
                     dp->xmin+(xoff+xscale)*(dp->xmax-dp->xmin)/(dp->xsize-1.),
                     dp->ymin+(yoff+yscale)*(dp->ymax-dp->ymin)/(dp->ysize-1.),
                      psx0+lx0,psy0,psx0+pagex*72.,psy0+pagey*72.,
                      fs,TKMAJ*AXMAR*ps72,AXMAR*(1.-TKMAJ)*ps72,"%g","%g");
        }
    theta=thrad*57.29577;
    xsquish=cos(thrad);
    ysquish=sin(thrad);
    def_colors();
    if (viewtype!='c')
        g_psout("/SETPATH {CL GS} def\n");
    else
        g_psout("/SETPATH {CL} def\n");
    g_psout("/P6 { MT RL RL RL RL RL SETPATH} def\n");
    g_psout("/P5 { MT RL RL RL RL SETPATH} def\n");
    g_psout("/P4 { MT RL RL RL SETPATH} def\n");
    g_psout("/P3 { MT RL RL SETPATH} def\n");
    if (viewtype=='c')
        for (i=3;i<=6;i++)
            for (j=0;j<(pscolor ? COLORSOH : GREYSOH);j++)
                {
                sprintf(buf,"/P%d%c {P%d COLOR%c fill} def\n",
                            i,j+'A',i,j+'A');
                g_psout(buf);
                }
    else
        for (i=3;i<=6;i++)
            for (j=0;j<COLORS3D;j++)
                {
                sprintf(buf,"/P%d%c {P%d COLOR%c fill GR COLOR%c stroke} def\n",
                            i,j*3+'A',i,j*3+'A',j*3+'C');
                g_psout(buf);
                sprintf(buf,"/P%d%c {P%d COLOR%c fill GR COLOR%c stroke} def\n",
                            i,j*3+'B',i,j*3+'B',j*3+'C');
                g_psout(buf);
                }
    }


static void set_pscolor(int layer,int type)

    {
    /* type = 0 top, 1 bot, 2 wire */
    char    buf[20];
    int     mod;

    mod=pscolor ? (viewtype=='c' ? COLORSOH : COLORS3D) :
                  (viewtype=='c' ? GREYSOH  : GREYS3D);
    if (viewtype!='c')
        mod=((layer+startlayer)%mod)*3+type+'A';
    else
        mod=((layer+startlayer)%mod)+'A';
    sprintf(buf,"COLOR%c ",mod);
    g_psout(buf);
    }


static void poly_type(int layer,int npoints,int top)

    {
    char    buf[20];
    int     mod;

    mod=pscolor ? (viewtype=='c' ? COLORSOH : COLORS3D) :
                  (viewtype=='c' ? GREYSOH  : GREYS3D);
    if (viewtype!='c')
        mod=((layer+startlayer)%mod)*3+'A'+ (top ? 0 : 1);
    else
        mod=((layer+startlayer)%mod)+'A';
    sprintf(buf,"P%d%c\n",npoints,mod);
    g_psout(buf);
    }


static void def_colors(void)

    {
    int     i;
    char    buf[50];

    if (pscolor)
        if (viewtype=='c')
            for (i=0;i<COLORSOH;i++)
                {
                sprintf(buf,"/COLOR%c {%g %g %g setrgbcolor} def\n",i+'A',
                        paletteOHc[i][0],paletteOHc[i][1],paletteOHc[i][2]);
                g_psout(buf);
                }
        else
            for (i=0;i<COLORS3D;i++)
                {
                sprintf(buf,"/COLOR%c {%g %g %g setrgbcolor} def\n",i*3+'A',
                        palette3Dc[i][0],palette3Dc[i][1],palette3Dc[i][2]);
                g_psout(buf);
                sprintf(buf,"/COLOR%c {%g %g %g setrgbcolor} def\n",i*3+'B',
                  palette3Dc[i][0]/2.,palette3Dc[i][1]/2.,palette3Dc[i][2]/1.5);
                g_psout(buf);
                sprintf(buf,"/COLOR%c {%g %g %g setrgbcolor} def\n",i*3+'C',
                  palette3Dc[i][0]/3.,palette3Dc[i][1]/3.,palette3Dc[i][2]/2.5);
                g_psout(buf);
                }
    else
        if (viewtype=='c')
            for (i=0;i<GREYSOH;i++)
                {
                sprintf(buf,"/COLOR%c {%g setgray} def\n",i+'A',paletteOHg[i]);
                g_psout(buf);
                }
        else
            for (i=0;i<GREYS3D;i++)
                {
                sprintf(buf,"/COLOR%c {%g setgray} def\n",i*3+'A',palette3Dg[i]);
                g_psout(buf);
                sprintf(buf,"/COLOR%c {%g setgray} def\n",i*3+'B',palette3Dg[i]);
                g_psout(buf);
                sprintf(buf,"/COLOR%c {%g setgray} def\n",i*3+'C',palette3Dg[i]/2.);
                g_psout(buf);
                }
    }




static void triangle3d(float x1,float y1,float z1,float x2,float y2,
                       float z2,float x3,float y3,float z3)

    {
    char    buf[100];
    int     i,j,k,top;
    float   xa[7],ya[7],za[7];
    int     xps[6],yps[6];
    float   px1,py1,px2,py2,px3,py3;
    float   zlo,zhi,zmax;

    zmax=zoff+zscale;
    if ((z1<zoff && z2<zoff && z3<zoff) || (z1>=zmax && z2>=zmax && z3>=zmax))
        return;

    px1=x3d(x1,y1); py1=y3d(x1,y1,z1);
    px2=x3d(x2,y2); py2=y3d(x2,y2,z2);
    px3=x3d(x3,y3); py3=y3d(x3,y3,z3);
    top = ((px1-px2)*(py3-py2)-(px3-px2)*(py1-py2)>0);
    for (i=0;i<nlayers;i++)
        {
        zlo=zboundary[i];
        zhi=zboundary[i+1];
        j=0;
        if ((z1<zlo && z2<zlo && z3<zlo) || (z1>=zhi && z2>=zhi && z3>=zhi))
            continue;
        clip_line(x1,y1,z1,x2,y2,z2,zlo,zhi,xa,ya,za,&j);
        clip_line(x2,y2,z2,x3,y3,z3,zlo,zhi,xa,ya,za,&j);
        clip_line(x3,y3,z3,x1,y1,z1,zlo,zhi,xa,ya,za,&j);
        for (k=0;k<j;k++)
            {
            xps[k] = 10.*x3d(xa[k],ya[k])+.5;
            yps[k] = 10.*y3d(xa[k],ya[k],za[k])+.5;
            }
        for (k=j-1;k>0;k--)
            {
            sprintf(buf,"%d %d ",xps[k]-xps[k-1],yps[k]-yps[k-1]);
            g_psout(buf);
            }
        sprintf(buf,"%d %d ",xps[0],yps[0]);
        g_psout(buf);
        poly_type(i,j,top);
        }
    }


static void tile3d(float x1,float y1,float z1,float x2,float y2,float z2,
                   float x3,float y3,float z3,float x4,float y4,float z4)

    {
    /* Order is important here */
    triangle3d(x1,y1,z1,x2,y2,z2,x3,y3,z3);
    triangle3d(x1,y1,z1,x3,y3,z3,x4,y4,z4);
    }


static float x3d(float x,float y)

    {
    return(gx0+gxx*x+gxy*y);
    }


static float y3d(float x,float y,float z)

    {
    return(gy0+gyx*x+gyy*y+gyz*z);
    }


static void line3d(float x1,float y1,float z1,float x2,float y2,float z2)

    {
    g_line(x3d(x1,y1),y3d(x1,y1,z1),x3d(x2,y2),y3d(x2,y2,z2));
    }


static void setupaxes(DATAPTR d,int grids)

    {
    char    buf[100];
    int     i,j,i1,i2,modval;
    float   offset,scale,val;
    float   min,max,pval;
    float   dx,dy,ts,z,x,ytext;

    if (viewtype=='c')
        {
        z_legend(d);
        ps_put_titles(d,(psleft+psright)/2.,pstop-fs,
                      x3d(xoff+xscale/2.,0.),
                      y3d(0.,yoff,0.)-ps72*AXMAR-fs*(1+XAXNMAR),
                      x3d(xoff,0.)-ps72*AXMAR-fs*(1+YAXNMAR),
                      y3d(0.,yoff+yscale/2.,0.));
        return;
        }
    g_grey(0.);
    g_dash(0.,0.);
    g_linejoin(0);
    g_linecap(2);
    g_linewidth(AXLW);

    /* X-AXIS */

    if (d->xsize<=1)
        {
        min=d->xmin;
        max=d->xmax;
        }
    else
        {
        min=d->xmin+(float)xoff*(d->xmax-d->xmin)/(d->xsize-1.);
        max=d->xmin+(float)(xoff+xscale)*(d->xmax-d->xmin)/(d->xsize-1.);
        }
    getticks(min,max,&offset,&scale,&i1,&i2,&modval,20);
    dx=-cos(thrad)*ps72*AXMAR;
    dy=-sin(thrad)*ps72*AXMAR;
    g_justify(1);
    for (i=i1;i<=i2;i++)
        {
        pval=offset+i*scale;
        if (fabs(pval/scale)<1e-4)
            pval=0;
        val=xindex(d,pval);
        if (nlayers>1)
            set_pscolor(0,0);
        else
            g_grey(0.);
        ts = ((i%modval)==0) ? TKMAJ : TKMIN;
        g_moveto(x3d(val,yoff),y3d(val,yoff,zoff));
        g_lineto(x3d(val,yoff)+dx*ts,y3d(val,yoff,zoff)+dy*ts);
        g_stroke();
        if (i%modval)
            continue;
        if (grids)
            {
            for (j=0;j<nlayers;j++)
                {
                if (nlayers>1)
                    set_pscolor(j,0);
                line3d(val,yoff+yscale,zboundary[j],
                       val,yoff+yscale,zboundary[j+1]);
                }
            if (nlayers>1)
                set_pscolor(0,0);
            line3d(val,yoff,zoff,val,yoff+yscale,zoff);
            }
        if (i>i1+1)
            {
            sprintf(buf,"%g",pval);
            g_moveto(x3d(val,yoff)+dx,y3d(val,yoff,zoff)+dy);
            if (viewtype=='a')
                g_slant(theta);
            else
                {
                g_rotate(-theta);
                g_slant(theta*2.);
                }
            g_grey(0.);
            g_showtext(buf);
            if (viewtype=='a')
                g_unslant(theta);
            else
                {
                g_unslant(theta*2.);
                g_rotate(theta);
                }
            }
        }
    /* X-axis label */
    g_justify(1);
    dx/=(ps72*AXMAR);
    dy/=(ps72*AXMAR);
    dx = dx * (ps72*AXMAR + fs*(1.+XAXNMAR));
    dy = dy * (ps72*AXMAR + fs*(1.+XAXNMAR));
    g_moveto(x3d(xoff+xscale/2.,yoff)+dx,y3d(xoff+xscale/2.,yoff,zoff)+dy);
    if (viewtype=='a')
        g_slant(theta);
    else
        {
        g_rotate(-theta);
        g_slant(2.*theta);
        }
    g_grey(0.);
    g_showtext(d->xtitle);
    if (viewtype=='a')
        g_unslant(theta);
    else
        {
        g_unslant(2.*theta);
        g_rotate(theta);
        }


    /* Y-AXIS */

    if (d->ysize<=1)
        {
        min=d->ymin;
        max=d->ymax;
        }
    else
        {
        min=d->ymin+(float)yoff*(d->ymax-d->ymin)/(d->ysize-1.);
        max=d->ymin+(float)(yoff+yscale)*(d->ymax-d->ymin)/(d->ysize-1.);
        }
    getticks(min,max,&offset,&scale,&i1,&i2,&modval,20);
    if (viewtype=='a')
        {
        dx=ps72*AXMAR;
        dy=0.;
        x=xoff+xscale;
        z=zoff;
        g_justify(1);
        }
    else
        {
        dx=cos(thrad)*ps72*AXMAR;
        dy=-sin(thrad)*ps72*AXMAR;
        x=xoff+xscale;
        z=zoff;
        g_justify(1);
        }
    for (i=i1;i<=i2;i++)
        {
        pval=offset+i*scale;
        if (fabs(pval/scale)<1e-4)
            pval=0;
        val=yindex(d,pval);
        if (nlayers>1)
            set_pscolor(0,0);
        else
            g_grey(0.);
        ts = ((i%modval)==0) ? TKMAJ : TKMIN;
        g_moveto(x3d(x,val),y3d(x,val,z));
        g_lineto(x3d(x,val)+dx*ts,y3d(x,val,z)+dy*ts);
        g_stroke();
        if (i%modval)
            continue;
        if (grids)
            {
            for (j=0;j<nlayers;j++)
                {
                if (nlayers>1)
                    set_pscolor(j,0);
                line3d(xoff,val,zboundary[j],xoff,val,zboundary[j+1]);
                }
            if (nlayers>1)
                set_pscolor(0,0);
            line3d(xoff,val,zoff,xoff+xscale,val,zoff);
            }
        if (i>i1+1)
            {
            sprintf(buf,"%g",pval);
            g_moveto(x3d(x,val)+dx,y3d(x,val,z)+dy);
            g_rotate(theta);
            if (viewtype=='a')
                g_slant(180.-theta);
            else
                g_slant(180.-2.*theta);
            g_grey(0.);
            g_showtext(buf);
            if (viewtype=='a')
                g_unslant(180.-theta);
            else
                g_unslant(180.-2.*theta);
            g_rotate(-theta);
            }
        }
    dx/=(ps72*AXMAR);
    dy/=(ps72*AXMAR);
    dx*=(ps72*AXMAR+fs*(1.+XAXNMAR));
    dy*=(ps72*AXMAR+fs*(1.+XAXNMAR));
    g_moveto(x3d(x,yoff+yscale/2.)+dx,y3d(x,yoff+yscale/2.,z)+dy);
    g_rotate(theta);
    if (viewtype=='a')
        g_slant(180.-theta);
    else
        g_slant(180.-2.*theta);
    g_grey(0.);
    g_showtext(d->ytitle);
    if (viewtype=='a')
        g_unslant(180.-theta);
    else
        g_unslant(180.-2.*theta);
    g_rotate(-theta);


    /* Z-AXIS */

    getticks(zoff,zoff+zscale,&offset,&scale,&i1,&i2,&modval,20);
    dx=-cos(thrad)*ps72*AXMAR;
    dy=-sin(thrad)*ps72*AXMAR;
    g_justify(5);
    if (nlayers<=1)
        {
        g_grey(0.);
        for (i=i1;i<=i2;i++)
            {
            z=offset+i*scale;
            if (fabs(z/scale)<1e-4)
                z=0;
            ts = ((i%modval)==0) ? TKMAJ : TKMIN;
            g_moveto(x3d(xoff,yoff),y3d(xoff,yoff,z));
            g_lineto(x3d(xoff,yoff)+dx*ts,y3d(xoff,yoff,z)+dy*ts);
            g_stroke();
            if (i%modval)
                continue;
            if (grids)
                {
                line3d(xoff,yoff+yscale,z,xoff,yoff,z);
                line3d(xoff,yoff+yscale,z,xoff+xscale,yoff+yscale,z);
                }
            if (i>i1+1)
                {
                sprintf(buf,"%g",z);
                g_moveto(x3d(xoff,yoff)+dx,y3d(xoff,yoff,z)+dy);
                g_rotate(theta);
                g_slant(90.-theta);
                g_showtext(buf);
                g_unslant(90.-theta);
                g_rotate(-theta);
                }
            }
        }
    else
        {
        ytext=-1000.;
        for (i=0;i<=nlayers;i++)
            {
            z=zboundary[i];
            if (grids)
                {
                set_pscolor(i==nlayers ? nlayers-1 : i,0);
                line3d(xoff,yoff+yscale,z,xoff,yoff,z);
                line3d(xoff,yoff+yscale,z,xoff+xscale,yoff+yscale,z);
                }
            set_pscolor(i==nlayers ? nlayers-1 : i,0);
            g_moveto(x3d(xoff,yoff),y3d(xoff,yoff,z));
            g_lineto(x3d(xoff,yoff)+dx*TKMAJ,y3d(xoff,yoff,z)+dy*TKMAJ);
            g_stroke();
            if (y3d(xoff,yoff,z)-ytext > fs*1.1)
                {
                sprintf(buf,"%g",z);
                g_moveto(x3d(xoff,yoff)+dx,y3d(xoff,yoff,z)+dy);
                g_rotate(theta);
                g_slant(90.-theta);
                g_showtext(buf);
                g_unslant(90.-theta);
                g_rotate(-theta);
                ytext=y3d(xoff,yoff,z);
                }
            if (i<nlayers)
                three_verts(grids,z,zboundary[i+1]);
            }
        }

    /* Draw 3-D outline box */

    g_grey(0.);
    g_linewidth(BOXLWIDTH);
    if (nlayers>1)
        set_pscolor(0,0);
    else
        {
        g_grey(0.);
        line3d(xoff,yoff,zoff,xoff,yoff,zoff+zscale);
        }
    line3d(xoff+xscale,yoff,zoff,xoff,yoff,zoff);
    line3d(xoff+xscale,yoff+yscale,zoff,xoff+xscale,yoff,zoff);
    if (grids)
        {
        line3d(xoff,yoff,zoff,xoff,yoff+yscale,zoff);
        line3d(xoff,yoff+yscale,zoff,xoff+xscale,yoff+yscale,zoff);
        if (nlayers>1)
            set_pscolor(nlayers-1,0);
        else
            {
            g_grey(0.);
            line3d(xoff,yoff+yscale,zoff,xoff,yoff+yscale,zoff+zscale);
            line3d(xoff+xscale,yoff+yscale,zoff,
                   xoff+xscale,yoff+yscale,zoff+zscale);
            }
        line3d(xoff,yoff,zoff+zscale,xoff,yoff+yscale,zoff+zscale);
        line3d(xoff,yoff+yscale,zoff+zscale,
               xoff+xscale,yoff+yscale,zoff+zscale);
        }

    /* Graph title */
    g_justify(1);
    g_grey(0.);
    g_moveto((psleft+psright)/2.,pstop);
    g_showtext(d->gtitle);

    /* Z-axis label */
    g_justify(7);
    dx=-(ps72*AXMAR+fs*(1.+YAXNMAR))*cos(thrad);
    dy=-(ps72*AXMAR+fs*(1.+YAXNMAR))*sin(thrad);
    g_moveto(x3d(xoff,yoff)+dx,y3d(xoff,yoff,zoff+zscale/2.)+dy);
    g_rotate(90.);
    g_slant(theta+90.);
    g_grey(0.);
    g_showtext(d->ztitle);
    g_unslant(theta+90.);
    g_rotate(-90.);
    }


static void three_verts(int grids,float z0,float z1)

    {
    g_linewidth(BOXLWIDTH);
    line3d(xoff,yoff,z0,xoff,yoff,z1);
    if (grids)
        {
        line3d(xoff,yoff+yscale,z0,xoff,yoff+yscale,z1);
        line3d(xoff+xscale,yoff+yscale,z0,xoff+xscale,yoff+yscale,z1);
        }
    g_linewidth(AXLW);
    }


static void z_legend(DATAPTR d)

    {
    int     i;
    float   ytext,x1,x2,yt,yb,y0,y1;
    float   a0,a1;
    char    buf[40];

    ytext=-100;
    x1=psx0;
    x2=psx0+LEGX*pagex*72.;
    yb=psy0;
    yt=psy0+pagey*72.;
    g_linewidth(AXLW);
    g_justify(5);
    for (i=0;i<=nlayers;i++)
        {
        a0=(zboundary[i]-zoff)/zscale;
        y0=y3d(xoff,yoff+yscale*a0,0.);
        if (i<nlayers)
            {
            a1=(zboundary[i+1]-zoff)/zscale;
            y1=y3d(xoff,yoff+yscale*a1,0.);
            set_pscolor(i,0);
            g_moveto(x1,y0);
            g_lineto(x2,y0);
            g_lineto(x2,y1);
            g_lineto(x1,y1);
            g_closepath();
            g_fill();
            }
        g_grey(0.);
        g_moveto(x1,y0);
        g_lineto(x1-ps72*AXMAR*TKMAJ,y0);
        g_stroke();
        if (y0-ytext>fs)
            {
            sprintf(buf,"%g",zboundary[i]);
            g_moveto(x1-ps72*AXMAR,y0);
            g_showtext(buf);
            ytext=y0;
            }
        }
    g_linewidth(BOXLWIDTH);
    g_grey(0.);
    g_moveto(x1,yb);
    g_lineto(x1,yt);
    g_lineto(x2,yt);
    g_lineto(x2,yb);
    g_closepath();
    g_stroke();
    g_justify(7);
    g_moveto(x3d(xoff,0.)-ps72*AXMAR-fs*(1.+YAXNMAR)-lx0,
             y3d(0.,yoff+yscale/2.,0.));
    g_rotate(90.);
    g_showtext(d->ztitle);
    g_rotate(-90.);
    }
